/*	$NetBSD$	*/

/*-
 * Copyright (c) 2010 SHIMIZU Ryo <ryo@nerv.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD$");

#define _INTR_PRIVATE	/* for arm/pic/picvar.h */

#include "locators.h"

#include <sys/param.h>
#include <sys/evcnt.h>
#include <sys/device.h>
#include <machine/bus.h>
#include <arm/cpufunc.h>
#include <arm/imx/imx51reg.h>
#include <arm/imx/imx51var.h>

static int tzic_match(device_t, cfdata_t, void *);
static void tzic_attach(device_t, device_t, void *);

/* for arm/pic */
static void tzic_unblock_irqs(struct pic_softc *, size_t, uint32_t);
static void tzic_block_irqs(struct pic_softc *, size_t, uint32_t);
static void tzic_establish_irq(struct pic_softc *, struct intrsource *);
static void tzic_source_name(struct pic_softc *, int, char *, size_t);

struct tzic_softc {
	device_t sc_dev;
	bus_addr_t sc_addr;
	bus_size_t sc_size;
	bus_space_tag_t sc_iot;
	bus_space_handle_t sc_ioh;
	struct pic_softc sc_pic;
	uint32_t sc_enabled_irq[4];	/* IRQ0-127 */
};

const struct pic_ops tzic_pic_ops = {
	.pic_unblock_irqs = tzic_unblock_irqs,
	.pic_block_irqs = tzic_block_irqs,
	.pic_establish_irq = tzic_establish_irq,
	.pic_source_name = tzic_source_name
};


void tzic_dump(void);
static void tzic_init(struct tzic_softc *);

static struct tzic_softc *tzic_sc;
static const char * const tzic_intr_source_names[] = TZIC_INTR_SOURCE_NAMES;

CFATTACH_DECL_NEW(tzic, sizeof(struct tzic_softc),
    tzic_match, tzic_attach, NULL, NULL);

int
tzic_match(device_t parent, cfdata_t self, void *aux)
{
	struct axi_attach_args *aa;

	aa = aux;

	if (aa->aa_addr != IMX51_TZIC_BASE)
		return 0;

	if (tzic_sc != NULL)
		return 0;	/* already attached */

	aa->aa_size = IMX51_TZIC_SIZE;
	return 1;
}

void
tzic_attach(device_t parent, device_t self, void *aux)
{
	struct tzic_softc *sc;
	struct axi_attach_args *aa;
	int i;

	aprint_normal(": TrustZone Interrupt Controller\n");
	aprint_naive("\n");

	aa = aux;
	tzic_sc = sc = device_private(self);
	sc->sc_dev = self;
	sc->sc_iot = aa->aa_iot;
	sc->sc_addr = aa->aa_addr;
	sc->sc_size = aa->aa_size;
	if (bus_space_map(sc->sc_iot, sc->sc_addr, sc->sc_size, 0, &sc->sc_ioh))
		panic("%s: Cannot map TZIC registers", device_xname(self));

	for (i = 0; i < 4; i++)
		sc->sc_enabled_irq[i] = 0;

	/* XXX: debug */
	tzic_dump();

	tzic_init(sc);

	sc->sc_pic.pic_ops = &tzic_pic_ops;
	sc->sc_pic.pic_maxsources = IMX51_TZIC_INTNUM;
	strlcpy(sc->sc_pic.pic_name, device_xname(self),
	    sizeof(sc->sc_pic.pic_name));

	pic_add(&sc->sc_pic, 0);

	cpsie(I32_bit);
}

void
tzic_dump()
{
	struct tzic_softc *sc;
	int i;

	sc = tzic_sc;

	printf("==== TZIC DUMP ====\n");
	printf("INTCTRL : %08x\n",
	    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_INTCTRL_REG));
	printf("INTTYPE : %08x\n",
	    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_INTTYPE_REG));
	printf("PRIOMASK: %08x\n",
	    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_PRIOMASK_REG));
	printf("SYNCCTRL: %08x\n",
	    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_SYNCCTRL_REG));
	printf("DSMINT  : %08x\n",
	    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_DSMINT_REG));
	for (i = 0; i < 4; i++)
		printf("INTSEC%d : (%d-%d) %08x\n", i, i * 32, i * 32 + 31,
		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_INTSEC_REG(i)));
	for (i = 0; i < 4; i++)
		printf("ENSET%d  : (%d-%d) %08x\n", i, i * 32, i * 32 + 31,
		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_ENSET_REG(i)));
	for (i = 0; i < 4; i++)
		printf("ENCLEAR%d: (%d-%d) %08x\n", i, i * 32, i * 32 + 31,
		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_ENCLEAR_REG(i)));
	for (i = 0; i < 4; i++)
		printf("SRCSET%d : (%d-%d) %08x\n", i, i * 32, i * 32 + 31,
		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_SRCSET_REG(i)));
	for (i = 0; i < 4; i++)
		printf("SRCCLEAR%d: (%d-%d) %08x\n", i, i * 32, i * 32 + 31,
		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_SRCCLEAR_REG(i)));
//	for (i = 0; i < 32; i++)
//		printf("PRIORITY%d: (%d-%d) %08x\n", i, i * 4, i * 4 + 3,
//		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_PRIORITY_REG(i)));
	for (i = 0; i < 4; i++)
		printf("PND%d   : (%d-%d) %08x\n", i, i * 32, i * 32 + 31,
		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_PND_REG(i)));
	for (i = 0; i < 4; i++)
		printf("HIPND%d : (%d-%d) %08x\n", i, i * 32, i * 32 + 31,
		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_HIPND_REG(i)));
//	for (i = 0; i < 4; i++)
//		printf("WAKEUP%d: (%d-%d) %08x\n", i, i * 32, i * 32 + 31,
//		    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_WAKEUP_REG(i)));
//	printf("SWINT  : %08x\n",
//	    bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_SWINT_REG));
}


static void
tzic_init(struct tzic_softc *sc)
{
	int i;

	/* enable interrupts */
	bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_INTCTRL_REG);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh,
	    IMX51_TZIC_INTCTRL_REG, 0x80010001);
	bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_INTCTRL_REG);

	/* set priority mask */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh,
	    IMX51_TZIC_PRIOMASK_REG, 0x0000001f);
	bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_PRIOMASK_REG);

	/* set latency mode */
	bus_space_write_4(sc->sc_iot, sc->sc_ioh,
	    IMX51_TZIC_SYNCCTRL_REG, 0x00000002);
	bus_space_read_4(sc->sc_iot, sc->sc_ioh, IMX51_TZIC_SYNCCTRL_REG);


	/* all interrupts are non-secure; no FIQ, use IRQ. */
	for (i = 0; i < 4; i++)
		bus_space_write_4(sc->sc_iot, sc->sc_ioh,
		    IMX51_TZIC_INTSEC_REG(i), 0xffffffff);

	/* disable all interrupts */
	for (i = 0; i < 4; i++)
		bus_space_write_4(sc->sc_iot, sc->sc_ioh,
		    IMX51_TZIC_ENCLEAR_REG(i), 0xffffffff);

#if 0
	/* set priority */
	for (i = 0; i < 32; i++)
		bus_space_write_4(sc->sc_iot, sc->sc_ioh,
		    IMX51_TZIC_PRIORITY_REG(i), 0xff);

	bus_space_write_4(sc->sc_iot, sc->sc_ioh,
	    IMX51_TZIC_DSMINT_REG, 0);			/* DeepSleepMode interrupt sync update */
#endif

	/* clear interrupt source */
//	for (i = 0; i < 4; i++)
//		bus_space_write_4(sc->sc_iot, sc->sc_ioh,
//		    IMX51_TZIC_SRCSET_REG(i), 0);
	for (i = 0; i < 4; i++)
		bus_space_write_4(sc->sc_iot, sc->sc_ioh,
		    IMX51_TZIC_SRCCLEAR_REG(i), 0xffffffff);

#if 0
	for (i = 0; i < 4; i++)
		bus_space_write_4(sc->sc_iot, sc->sc_ioh,	/* XXX: cannot write */
		    IMX51_TZIC_PND_REG(i), 0);
#endif
#if 0
	for (i = 0; i < 4; i++)
		bus_space_write_4(sc->sc_iot, sc->sc_ioh,	/* XXX: cannot write */
		    IMX51_TZIC_HIPND_REG(i), 0);
#endif

#if 0
	for (i = 0; i < 4; i++)
		bus_space_write_4(sc->sc_iot, sc->sc_ioh,	/* no interrupt assert wakeup */
		    IMX51_TZIC_WAKEUP_REG(i), 0x00000000);

//	bus_space_write_4(sc->sc_iot, sc->sc_ioh,
//	    IMX51_TZIC_SWINT_REG, 0);

	bus_space_write_4(sc->sc_iot, sc->sc_ioh,
	    IMX51_TZIC_INTTYPE_REG, 0x00000003);
#endif

#if 0
	/* XXX: DEBUG: occur software interrupt */
	bus_space_write_4(tzic_sc->sc_iot, tzic_sc->sc_ioh,
	    IMX51_TZIC_ENSET_REG(1), 0x00000100);
	bus_space_write_4(tzic_sc->sc_iot, tzic_sc->sc_ioh,
	    IMX51_TZIC_ENSET_REG(1), 0x00000200);

	bus_space_write_4(sc->sc_iot, sc->sc_ioh,
	    IMX51_TZIC_SWINT_REG, IRQ_EPIT1);
	bus_space_write_4(sc->sc_iot, sc->sc_ioh,
	    IMX51_TZIC_SWINT_REG, IRQ_EPIT2);
#endif
}

void
tzic_unblock_irqs(struct pic_softc *pic, size_t irq_base, uint32_t irq_mask)
{
	int n;

	KASSERT(irq_base < IMX51_TZIC_INTNUM);

	n = irq_base / 32;
	tzic_sc->sc_enabled_irq[n] |= irq_mask;

	printf("%s:%d: irq_base=%ld: ENSET[%d]=%08x\n", __func__, __LINE__, irq_base, n, irq_mask);
	printf("%s:%d: sc_enabled_irq[%d] = 0x%08x\n", __func__, __LINE__, n, tzic_sc->sc_enabled_irq[n]);

	bus_space_write_4(tzic_sc->sc_iot, tzic_sc->sc_ioh,
	    IMX51_TZIC_ENSET_REG(n), irq_mask);
}

void
tzic_block_irqs(struct pic_softc *pic, size_t irq_base, uint32_t irq_mask)
{
	int n;

	KASSERT(irq_base < IMX51_TZIC_INTNUM);

	n = irq_base / 32;
	tzic_sc->sc_enabled_irq[n] &= ~irq_mask;

	printf("%s:%d: irq_base=%ld: ENCLEAR[%d]=%08x\n", __func__, __LINE__, irq_base, n, irq_mask);
	printf("%s:%d: sc_enabled_irq[%d] = 0x%08x\n", __func__, __LINE__, n, tzic_sc->sc_enabled_irq[n]);

	bus_space_write_4(tzic_sc->sc_iot, tzic_sc->sc_ioh,
	    IMX51_TZIC_ENCLEAR_REG(n), irq_mask);
	bus_space_write_4(tzic_sc->sc_iot, tzic_sc->sc_ioh,
	    IMX51_TZIC_SRCCLEAR_REG(n), irq_mask);
}

void
tzic_establish_irq(struct pic_softc *pic, struct intrsource *is)
{
	uint32_t irq_offset, irq_mask;
	int n;

	irq_offset = is->is_irq & 31;
	irq_mask = 1 << irq_offset;

	n = is->is_irq / 32;

	printf("%s:%d: irq=%d: ENSET[%d]=%08x\n", __func__, __LINE__, is->is_irq, n, irq_mask);

	bus_space_write_4(tzic_sc->sc_iot, tzic_sc->sc_ioh,
	    IMX51_TZIC_ENSET_REG(n), irq_mask);
}

void
tzic_source_name(struct pic_softc *pic, int irq, char *buf, size_t len)
{
	strlcpy(buf, tzic_intr_source_names[irq], len);
}

void
imx51_irq_handler(void *frame)
{
	struct cpu_info *ci;
	struct pic_softc *pic;
	uint32_t oldipl_mask, pending;
	int oldipl, n;
	int ipl_mask;

	pic = &tzic_sc->sc_pic;

	ci = curcpu();
	oldipl = ci->ci_cpl;
	oldipl_mask = __BIT(oldipl);
	ipl_mask = 0;

	printf("%s:%d: ipl=%d\n", __func__, __LINE__, oldipl);

	for (n = 0; n < 4; n++) {
		pending = bus_space_read_4(tzic_sc->sc_iot, tzic_sc->sc_ioh,
		    IMX51_TZIC_SRCSET_REG(n));
		bus_space_write_4(tzic_sc->sc_iot, tzic_sc->sc_ioh,
		    IMX51_TZIC_SRCCLEAR_REG(n), pending);

		pending &= tzic_sc->sc_enabled_irq[n];
		if (pending == 0)
			continue;

		ipl_mask |= pic_mark_pending_sources(pic, n * 32, pending);

		printf("%s:%d: irqbase=%d, pending=0x%08x\n",
		    __func__, __LINE__, n * 32, pending);
	}

	if ((ipl_mask & ~oldipl_mask) > oldipl_mask)
		pic_do_pending_ints(I32_bit, oldipl, frame);

	printf("%s:%d: oldipl=%d\n", __func__, __LINE__, oldipl);
}
